@@ -1,6 +1,8 @@
 import logging
+import os
 import types
 import unittest
+import warnings
 from itertools import product, repeat
 from multiprocessing import Barrier, Value
 from unittest.mock import patch
@@ -10,6 +12,7 @@ from tqdm import tqdm
 
 from mpire import cpu_count, WorkerPool
 from mpire.context import FORK_AVAILABLE, RUNNING_WINDOWS
+from mpire.tqdm_utils import TqdmLock
 
 # Skip start methods that use fork if it's not available
 if not FORK_AVAILABLE:
@@ -22,6 +25,10 @@ def square(idx, x):
     return idx, x * x
 
 
+def extremely_large_output(idx, x):
+    return idx, os.urandom(1024 * 1024)
+
+
 def square_numpy(x):
     return x * x
 
@@ -53,7 +60,6 @@ class MapTest(unittest.TestCase):
             yield from iterable
 
         # Test results for different parameter settings
-        print()
         for n_jobs, n_tasks_max_active, worker_lifespan, chunk_size, n_splits in tqdm([
             (None, None, None, None, None),
             (1, None, None, None, None),
@@ -76,7 +82,8 @@ class MapTest(unittest.TestCase):
                                       chunk_size=chunk_size, n_splits=n_splits):
 
                         # Test if parallel map results in the same as ordinary map function. Should work both for
-                        # generators and iterators. Also check if an empty list works as desired.
+                        # generators and iterators. Also check if an empty list and extremely large output (exceeding
+                        # os.pipe limits) works as desired.
                         results_list = map_func(square, self.test_data, max_tasks_active=n_tasks_max_active,
                                                 worker_lifespan=worker_lifespan)
                         self.assertIsInstance(results_list, result_type)
@@ -102,6 +109,16 @@ class MapTest(unittest.TestCase):
                         self.assertIsInstance(results_list, result_type)
                         self.assertEqual([], list(results_list))
 
+                    # When the os pipe capacity is exceeded, a worker restart based on worker lifespan would hang if we
+                    # not fetch all the results from a worker. We only verify the amount of data returned here.
+                    with self.subTest(map_func=map_func, output='data exceeding pipe limits', n_jobs=n_jobs,
+                                      n_tasks_max_active=n_tasks_max_active, worker_lifespan=worker_lifespan,
+                                      chunk_size=chunk_size, n_splits=n_splits):
+                        results_list = map_func(extremely_large_output, self.test_data,
+                                                max_tasks_active=n_tasks_max_active, worker_lifespan=worker_lifespan)
+                        self.assertIsInstance(results_list, result_type)
+                        self.assertEqual(len(self.test_desired_output), len(list(results_list)))
+
     def test_numpy_input(self):
         """
         Test map with numpy input
@@ -721,7 +738,7 @@ class CPUPinningTest(unittest.TestCase):
                 if cpu_ids is None:
                     self.assertEqual(p.call_args_list, [])
                 else:
-                    self.assertEqual(p.call_count, pool.params.n_jobs)
+                    self.assertEqual(p.call_count, pool.pool_params.n_jobs)
                     mask = [call[0][1] for call in p.call_args_list]
                     self.assertListEqual(mask, expected_mask)
 
@@ -747,7 +764,7 @@ class CPUPinningTest(unittest.TestCase):
 
                 # Verify that CPU pinning is used as many times as there are jobs and is called for each worker process
                 # ID
-                self.assertEqual(p.call_count, pool.params.n_jobs)
+                self.assertEqual(p.call_count, pool.pool_params.n_jobs)
                 mask = [call[0][1] for call in p.call_args_list]
                 self.assertListEqual(mask, expected_mask)
 
@@ -775,6 +792,15 @@ class ProgressBarTest(unittest.TestCase):
         self.test_desired_output_numpy = square_numpy(self.test_data_numpy)
         self.test_data_len_numpy = len(self.test_data_numpy)
 
+        # Get original tqdm lock
+        self.original_tqdm_lock = tqdm.get_lock()
+
+    def tearDown(self):
+        # The TQDM lock is temporarily changed when using a progress bar in MPIRE, here we check if it is restored
+        # correctly afterwards.
+        self.assertNotIsInstance(tqdm.get_lock(), TqdmLock)
+        self.assertEqual(tqdm.get_lock(), self.original_tqdm_lock)
+
     def test_valid_progress_bars_regular_input(self):
         """
         Valid progress bars are either False/True
@@ -833,7 +859,7 @@ class ProgressBarTest(unittest.TestCase):
                 # Progress bar on Windows with threading is currently not supported
                 if RUNNING_WINDOWS and start_method == 'threading':
                     with self.assertRaises(ValueError):
-                        results_list = pool.map(square, self.test_data, progress_bar=True)
+                        pool.map(square, self.test_data, progress_bar=True)
                 else:
                     results_list = pool.map(square, self.test_data, progress_bar=True)
                     self.assertIsInstance(results_list, list)
@@ -912,170 +938,41 @@ class KeepAliveTest(unittest.TestCase):
                                      self.test_desired_output_f1)
                 self.assertEqual(counter.value, n_jobs)
 
-    def test_keep_alive_func_changes(self):
+    def test_keep_alive_map_params_change(self):
         """
-        When keep_alive is set to True it should reuse existing workers between map calls, but only when the called
-        function, amongst others, is kept constant
+        When keep_alive is set to True it should reuse existing workers between map calls, even when the called
+        function, init or exit functions, or the worker lifespan changes
         """
         for n_jobs in [1, 3]:
             barrier = Barrier(n_jobs)
             counter = Value('i', 0)
             shared = barrier, counter
-            with self.subTest(n_jobs=n_jobs), \
+            with self.subTest(n_jobs=n_jobs), warnings.catch_warnings(), \
                     WorkerPool(n_jobs=n_jobs, shared_objects=shared, use_worker_state=True, keep_alive=True) as pool:
 
-                self.assertListEqual(pool.map(self._f1, self.test_data, worker_init=self._init1),
+                warnings.simplefilter('ignore')
+                self.assertListEqual(pool.map(self._f1, self.test_data, worker_lifespan=100, worker_init=self._init1,
+                                              worker_exit=self._exit1),
                                      self.test_desired_output_f1)
                 self.assertEqual(counter.value, n_jobs)
                 barrier.reset()
 
-                self.assertListEqual(list(pool.imap(self._f2, self.test_data, worker_init=self._init1)),
+                self.assertListEqual(list(pool.imap(self._f2, self.test_data, worker_lifespan=100,
+                                                    worker_init=self._init1, worker_exit=self._exit2)),
                                      self.test_desired_output_f2)
-                self.assertEqual(counter.value, n_jobs * 2)
+                self.assertEqual(counter.value, n_jobs)
                 barrier.reset()
 
-                self.assertListEqual(pool.map(self._f2, self.test_data, worker_init=self._init1),
+                self.assertListEqual(pool.map(self._f2, self.test_data, worker_lifespan=200, worker_init=self._init2,
+                                              worker_exit=self._exit1),
                                      self.test_desired_output_f2)
-                self.assertEqual(counter.value, n_jobs * 2)
-                barrier.reset()
-
-                self.assertListEqual(pool.map(self._f1, self.test_data, worker_init=self._init1),
-                                     self.test_desired_output_f1)
-                self.assertEqual(counter.value, n_jobs * 3)
-
-    def test_keep_alive_worker_lifespan_changes(self):
-        """
-        When keep_alive is set to True it should reuse existing workers between map calls, but only when the worker
-        lifespan, amongst others, is kept constant
-        """
-        for n_jobs in [1, 3]:
-            barrier = Barrier(n_jobs)
-            counter = Value('i', 0)
-            shared = barrier, counter
-            with self.subTest(n_jobs=n_jobs), \
-                    WorkerPool(n_jobs=n_jobs, shared_objects=shared, use_worker_state=True, keep_alive=True) as pool:
-
-                self.assertListEqual(pool.map(self._f1, self.test_data, worker_lifespan=100, worker_init=self._init1),
-                                     self.test_desired_output_f1)
                 self.assertEqual(counter.value, n_jobs)
                 barrier.reset()
 
-                self.assertListEqual(list(pool.imap(self._f1, self.test_data, worker_lifespan=100,
-                                                    worker_init=self._init1)), self.test_desired_output_f1)
-                self.assertEqual(counter.value, n_jobs)
-                barrier.reset()
-
-                self.assertListEqual(pool.map(self._f1, self.test_data, worker_lifespan=200, worker_init=self._init1),
-                                     self.test_desired_output_f1)
-                self.assertEqual(counter.value, n_jobs * 2)
-                barrier.reset()
-
-                self.assertListEqual(pool.map(self._f1, self.test_data, worker_lifespan=200, worker_init=self._init1),
-                                     self.test_desired_output_f1)
-                self.assertEqual(counter.value, n_jobs * 2)
-                barrier.reset()
-
-                self.assertListEqual(pool.map(self._f1, self.test_data, worker_lifespan=100, worker_init=self._init1),
-                                     self.test_desired_output_f1)
-                self.assertEqual(counter.value, n_jobs * 3)
-
-    def test_keep_alive_worker_init_changes(self):
-        """
-        When keep_alive is set to True it should reuse existing workers between map calls, but only when the worker init
-        function, amongst others, is kept constant
-        """
-        for n_jobs in [1, 3]:
-            barrier = Barrier(n_jobs)
-            counter = Value('i', 0)
-            shared = barrier, counter
-            with self.subTest(n_jobs=n_jobs), \
-                    WorkerPool(n_jobs=n_jobs, shared_objects=shared, use_worker_state=True, keep_alive=True) as pool:
-
-                self.assertListEqual(pool.map(self._f1, self.test_data, worker_init=self._init1),
+                self.assertListEqual(pool.map(self._f1, self.test_data, worker_lifespan=100, worker_init=self._init1,
+                                              worker_exit=None),
                                      self.test_desired_output_f1)
                 self.assertEqual(counter.value, n_jobs)
-                barrier.reset()
-
-                self.assertListEqual(list(pool.imap(self._f1, self.test_data, worker_init=self._init2)),
-                                     self.test_desired_output_f1)
-                self.assertEqual(counter.value, n_jobs * 2)
-                barrier.reset()
-
-                self.assertListEqual(pool.map(self._f1, self.test_data, worker_init=self._init2),
-                                     self.test_desired_output_f1)
-                self.assertEqual(counter.value, n_jobs * 2)
-                barrier.reset()
-
-                self.assertListEqual(pool.map(self._f1, self.test_data, worker_init=self._init1),
-                                     self.test_desired_output_f1)
-                self.assertEqual(counter.value, n_jobs * 3)
-
-    def test_keep_alive_worker_exit_changes(self):
-        """
-        When keep_alive is set to True it should reuse existing workers between map calls, but only when the worker exit
-        function, amongst others, is kept constant
-        """
-        for n_jobs in [1, 3]:
-            barrier = Barrier(n_jobs)
-            counter = Value('i', 0)
-            shared = barrier, counter
-            with self.subTest(n_jobs=n_jobs), \
-                    WorkerPool(n_jobs=n_jobs, shared_objects=shared, use_worker_state=True, keep_alive=True) as pool:
-
-                self.assertListEqual(pool.map(self._f1, self.test_data, worker_init=self._init1,
-                                              worker_exit=self._exit1), self.test_desired_output_f1)
-                self.assertEqual(counter.value, n_jobs)
-                barrier.reset()
-
-                self.assertListEqual(list(pool.imap(self._f1, self.test_data, worker_init=self._init1,
-                                                    worker_exit=self._exit2)), self.test_desired_output_f1)
-                self.assertEqual(counter.value, n_jobs * 2)
-                barrier.reset()
-
-                self.assertListEqual(pool.map(self._f1, self.test_data, worker_init=self._init1,
-                                              worker_exit=self._exit2), self.test_desired_output_f1)
-                self.assertEqual(counter.value, n_jobs * 2)
-                barrier.reset()
-
-                self.assertListEqual(pool.map(self._f1, self.test_data, worker_init=self._init1,
-                                              worker_exit=self._exit1), self.test_desired_output_f1)
-                self.assertEqual(counter.value, n_jobs * 3)
-
-    def test_keep_alive_enable_insights_changes(self):
-        """
-        When keep_alive is set to True it should reuse existing workers between map calls, but only when
-        enable_insights, amongst others, is kept constant
-        """
-        for n_jobs in [1, 3]:
-            barrier = Barrier(n_jobs)
-            counter = Value('i', 0)
-            shared = barrier, counter
-            with self.subTest(n_jobs=n_jobs), \
-                    WorkerPool(n_jobs=n_jobs, shared_objects=shared, use_worker_state=True, keep_alive=True) as pool:
-
-                self.assertListEqual(pool.map(self._f1, self.test_data, worker_init=self._init1, enable_insights=False),
-                                     self.test_desired_output_f1)
-                self.assertEqual(counter.value, n_jobs)
-                barrier.reset()
-
-                self.assertListEqual(list(pool.imap(self._f1, self.test_data, worker_init=self._init1,
-                                                    enable_insights=False)), self.test_desired_output_f1)
-                self.assertEqual(counter.value, n_jobs)
-                barrier.reset()
-
-                self.assertListEqual(pool.map(self._f1, self.test_data, worker_init=self._init1, enable_insights=True),
-                                     self.test_desired_output_f1)
-                self.assertEqual(counter.value, n_jobs * 2)
-                barrier.reset()
-
-                self.assertListEqual(pool.map(self._f1, self.test_data, worker_init=self._init1, enable_insights=True),
-                                     self.test_desired_output_f1)
-                self.assertEqual(counter.value, n_jobs * 2)
-                barrier.reset()
-
-                self.assertListEqual(pool.map(self._f1, self.test_data, worker_init=self._init1, enable_insights=False),
-                                     self.test_desired_output_f1)
-                self.assertEqual(counter.value, n_jobs * 3)
 
     def test_start_methods(self):
         """
@@ -1162,9 +1059,17 @@ class ExceptionTest(unittest.TestCase):
         self.logger = logging.getLogger()
         self.logger.setLevel(logging.DEBUG)
 
+        # Get original tqdm lock
+        self.original_tqdm_lock = tqdm.get_lock()
+
     def tearDown(self):
         self.logger.setLevel(logging.NOTSET)
 
+        # The TQDM lock is temporarily changed when using a progress bar in MPIRE, here we check if it is restored
+        # correctly afterwards.
+        self.assertNotIsInstance(tqdm.get_lock(), TqdmLock)
+        self.assertEqual(tqdm.get_lock(), self.original_tqdm_lock)
+
     def test_exceptions(self):
         """
         Tests if MPIRE can handle exceptions well
